/*
 * HorzBarCategory.java
 */
package UTEStatsViewer;
import com_ibm_us_bolmar.GraphicPlotter.*;
import java.awt.*;
import java.awt.event.*;
/**
 * The category of graphic elements used by the application.  Each
 * graphic element is a horizontal bar whose user coordinate width is
 * of type <tt>long</tt> and whose user coordinate height is of type
 * <tt>int</tt>.  An object of this class acts as a factor of the
 * graphic elements belonging to the category represented by the object.
 */
class HorzBarCategory extends Component implements MouseListener {
   /** The Component to which messages are written using setText(). */
   Label msgArea;

   // The name used when drawing this object.
   private String shortName;
   /** Return the shortName of the category. */
   public final String getShortName() {return shortName;}

   // The number of events represented by the category.  Each graphic
   // element represents one event.  Each accumulative graphic element
   // represents as many events as accumulations have been done into it.
   private long eventCnt;

   // The total amount of time represented by the category.
   private long widthSum;

   // The size of the horizontal bar painted for this object
   private final static int SAMPLE_BAR_WIDTH  = 34;
   private final static int SAMPLE_BAR_HEIGHT = 21;

   // The separation between the horizontal bar drawn for this
   // object and the name.
   private int nameSep;

   // The font metrics of the font used to draw this object.
   private FontMetrics fontMetrics;

   // Head of list of graphic elements in this category.
   private HorzBar head;

   // Parent in the category hierarchy.
   private HorzBarCategory parentCategory;

   /** Return whether category has a parent in the category hierarchy. */
   boolean hasParentCategory() {return (parentCategory != null);}

   // Sibling in the category hierarchy.
   private HorzBarCategory sibCategory;

   // Child in the category hierarchy.
   private HorzBarCategory childCategory;

   /** Return whether category has a child in the category hierarchy. */
   boolean hasChildCategory() {return (childCategory != null);}


   /** Hide default constructor. */
   private HorzBarCategory() {}


   /**
    * A category is defined by the following parameters.
    * @param name The name of the category used in messages.
    * @param shortName The name of the category used in the image drawn
    *    by paint().
    * @param color The color used in the image drawn by paint().
    * @param barHeight The height (in user coordinates) of the bars.
    * @param msgArea Where messages are written to bu using setText().
    */
   HorzBarCategory(String name, String shortName, Color color,
         Label msgArea) {
      setName(name);
      this.shortName = shortName;
      setForeground(color);
      this.msgArea = msgArea;
      addMouseListener(this);
   }


   /**
    * Constructor used when the name displayed in messages and the name
    * drawn as part of the category object are the same.
    */
   HorzBarCategory(String name, Color color, Label msgArea) {
      this(name, name, color, msgArea);
   }


   /**
    * Return the subcategory with the given name.  Create subcategory if
    * it does not exist.  The subcategories are ordered by name within
    * their parent category.
    * @param catetory parent category of the subcategory
    * @param name     name of the subcategory
    * @param color    color to use to draw subcategory
    */
   HorzBarCategory findAddSubcategory(String name, Color color)
   {
      HorzBarCategory sib;
      synchronized (this) {
         HorzBarCategory sibPrev = null;
         sib = childCategory;
         while (sib != null) {
            int cmprResult = name.compareTo(sib.getName());
            if (cmprResult > 0) {
               // subcategory ordered after current sibling
               sibPrev = sib;
               sib = sib.sibCategory;
            } else if (cmprResult == 0) {
               // subcategory already exists
               return sib;
            } else {
               // subcategory with name does not exist
               break;
            }
         }

         // insert new subcategory after sibPrev
         sib = new HorzBarCategory(name, color, msgArea);
         sib.parentCategory = this;
         sib.setVisible((childCategory == null) ? false :
                childCategory.isVisible());
         Component insertAfter;
         if (sibPrev == null) {
            // first subcategory under the category
            sib.sibCategory = childCategory;
            childCategory = sib;
            insertAfter = this;
         } else {
            sib.sibCategory = sibPrev.sibCategory;
            sibPrev.sibCategory = sib;
            insertAfter = sibPrev;
         }

         // add new subcategory to container in correct position
         Container cont = getParent();
         if (cont != null) {
            int len = cont.getComponentCount();
            int pos;
            for (pos = 0; pos < len; pos++) {
               if (cont.getComponent(pos) == insertAfter) {
                  break;
               }
            }
            cont.add(sib, Math.min(pos+1, len));
         }
      }
      return sib;
   }


   /**
    * Create a graphic element of this category.
    * @param xPos  x-position of the graphic element
    * @param yPos  y-position of the graphic element
    * @param width width of the graphic element
    * @param name  string to be included in messages about the graphic
    *       element
    */
   HorzBar createGraphic(long xPos, int yPos, long width, String name) {
      HorzBar ret = new HorzBar(xPos, yPos, width, name, this);
      accumulateStatistics(width);

      // add graphic element to list under this category
      ret.next = head;
      head = ret;

      return ret;
   }


   /** Return the preferred AWT size of the object. */
   public Dimension getPreferredSize() {
      // size consists of the size of the sample bar
      int width  = SAMPLE_BAR_WIDTH;
      int height = SAMPLE_BAR_HEIGHT;

      // plus the size of the name and space above and below the name
      fontMetrics = getFontMetrics(getFont());
      nameSep = 3 * fontMetrics.getLeading() / 2;
      height += fontMetrics.getHeight() + 2 * nameSep;
      width = Math.max(width, fontMetrics.stringWidth(shortName));

      return new Dimension(width, height);
   }


   /** Return the minimum AWT size of the object. */
   public Dimension getMinimumSize() {
      return getPreferredSize();
   }


   /** Paint the object. */
   public void paint(Graphics g) {
      int xName, xRect;

      // determine x-coordinate of rectangle and name so they are
      // centered with respect to each other
      int nameWidth = fontMetrics.stringWidth(shortName);
      if (nameWidth > SAMPLE_BAR_WIDTH) {
         xName = 0;
         xRect = (nameWidth - SAMPLE_BAR_WIDTH) / 2;
      } else {
         xRect = 0;
         xName = (SAMPLE_BAR_WIDTH - nameWidth) / 2;
      }

      // draw a filled rectangle with a border around it
      g.setColor(getForeground());
      g.fillRect(xRect + 1, 1, SAMPLE_BAR_WIDTH - 2,
            SAMPLE_BAR_HEIGHT - 2);
      g.setColor(Color.black);
      g.drawRect(xRect, 0, SAMPLE_BAR_WIDTH - 1, SAMPLE_BAR_HEIGHT - 1);

      // draw the name
      g.drawString(shortName, xName,
            SAMPLE_BAR_HEIGHT + nameSep + fontMetrics.getAscent());
   }


   /**
    * Paint the graphic elements of this category.
    * @param immediately if true, each graphic element is painted directly,
    *   instead of having repaint() invoked on it
    */
   private void paintGraphics(boolean immediately) {
      // recursively paint graphic elements in the subcatories
      for (HorzBarCategory category = childCategory; category != null;
            category = category.sibCategory) {
         category.paintGraphics(immediately);
      }

      for (HorzBar a = head; a != null; a = a.next) {
         if (immediately) {
            Graphics g = a.getGraphics();
            if (g != null) {
               try {
                  a.paint(g, 0, 0);  // note: application uses only depth 0
               } finally {
                  g.dispose();
               }
            }
         } else {
            a.repaint();
         }
      }
   }


   /**
    * Override setForeground to paint this category object and the
    * graphic elements in the category.
    */
   public void setForeground(Color c) {
      super.setForeground(c);

      // repaint the category icon; don't directly paint it because
      // the default AWT code fails to clip the getGraphics() for
      // a component to be within the clip region of the Container
      repaint();

      // directly paint the graphic elements
      paintGraphics(true);
   }


   /**
    * Invalidate this object and validate the top-level container that
    * it is in.
    */
   private final void revalidate() {
      invalidate();
      Container parent, nextParent;
      parent = this.getParent();
      if (parent != null) {
         while ((nextParent = parent.getParent()) != null) {
            parent = nextParent;
         }
         parent.validate();
      }
   }


   /**
    * Condense the category one level of detail.
    */
   void condenseCategoryHierarchy() {
      // condense the category only if it can be
      if (isVisible() && (parentCategory != null)) {
         // change visiblity of the category objects
         parentCategory.setVisible(true);
         for (HorzBarCategory category = parentCategory.childCategory;
               category != null; category = category.sibCategory) {
            category.setVisible(false);
         }
         paintGraphics(false);
         revalidate();
      }
   }


   /**
    * Expand the category one level of detail.
    */
   void expandCategoryHierarchy() {
      // expand the category only if it can be expanded
      if (isVisible() && (childCategory != null)) {
         // change visiblity of the category objects
         setVisible(false);
         for (HorzBarCategory category = childCategory; category != null;
               category = category.sibCategory) {
            category.setVisible(true);
         }
         paintGraphics(false);
         revalidate();
      }
   }


   /** Accumulate the statistics of a category. */
   final void accumulateStatistics(long width) {
      //  accumulate at each level of the category hierarchy
      for (HorzBarCategory category = this; category != null;
            category = category.parentCategory) {
         category.eventCnt++;
         category.widthSum += width;
      }
   }


   /**
    * Return the base category in the category hierarchy of this category.
    */
   final HorzBarCategory getBaseCategory() {
      HorzBarCategory current, parent;
      current = this;
      while ((parent = current.parentCategory) != null) {
         current = parent;
      }
      return current;
   }


   /**
    * Return the category in the category hierarchy that is the one
    * currently selected to be visible.
    */
   final HorzBarCategory getSelectedCategory() {
      // ascend category hierarchy until visible category is found
      HorzBarCategory category = this;
      while ((category != null) && !category.isVisible()) {
         category = category.parentCategory;
      }
      if (category == null) {
         // no visible category?
         category = this;
      }
      return category;
   }


   // Implement the MouseListener interface.
   public void mouseEntered(MouseEvent e) {}
   public void mouseExited(MouseEvent e) {}
   public void mouseClicked(MouseEvent e) {}

   public void mousePressed(MouseEvent e) {
      // workaround WIN32 implementation bug that sometimes dispatches
      // mouse events to the last lightweight component that the mouse
      // was on
      if (! contains(e.getX(), e.getY())) {
         return;
      }

      if ((e.getModifiers() & InputEvent.BUTTON3_MASK) != 0) {
         // right mouse button pressed, show color dialog

         // determine Frame that dialog is to be attached to
         Container topmostAncestor = null, nextParent;
         nextParent = this.getParent();
         while (nextParent != null) {
            topmostAncestor = nextParent;
            nextParent = nextParent.getParent();
         }

         (new HorzBarCategoryDialog((Frame) topmostAncestor, this)).show();
      } else {
         // show statistics
         StringBuffer buf = new StringBuffer(getName());
         buf.append(' ');
         buf.append(Message.get(Message.sum));
         buf.append('=');
         buf.append(TimeValue.toString(widthSum));
         buf.append(','); buf.append(' ');
         buf.append(Message.get(Message.count));
         buf.append('=');
         buf.append(eventCnt);
         if (eventCnt > 0) {
            buf.append(','); buf.append(' ');
            buf.append(Message.get(Message.average));
            buf.append('=');
            buf.append(TimeValue.toString(widthSum/eventCnt));
         }
         msgArea.setText(buf.toString());
      }
   }

   public void mouseReleased(MouseEvent e) {}
}
